#include <vector>
#include <map>

#include "FEM.h"

using namespace std;

//
void assembleLocalStiffMatrix ( const TriangMesh& triangMesh, int ElemIdx, double LocalMatrix[3][3] ) {
    // set zero
    for ( int i = 0; i < 3; ++i )
      for ( int j = 0; j < 3; ++j )
        LocalMatrix[i][j] = 0.;
      
    // get vertex positions
    std::vector< Vector > coords(3);
    for( int i = 0; i < 3; i++ )
      triangMesh.getVertexCoord( triangMesh.getGlobalIdx( ElemIdx, i), coords[i] );
    
    // gradPhi_i = e_i^\perp / (2A)
    // gradPhi_i * gradPhi_j = ( e_i * e_j ) / (4A^2)
    std::vector< Vector > gradPhi;
    for( int i = 0; i < 3; i++ ){
      // e_i = e_{i+2} - e_{i+1}
      gradPhi.push_back( coords[ (i+2)%3 ] );
      gradPhi[i] -= coords[ (i+1)%3 ];
    }

    // \int_T gradPhi_i * gradPhi_j = A * ( e_i * e_j ) / (4A^2)
    double vol = triangMesh.getArea ( ElemIdx );
    for ( int i = 0; i < 3; ++i )
      for ( int j = i; j < 3; ++j )
        LocalMatrix[j][i] = LocalMatrix[i][j] = 0.25 * (gradPhi[i] * gradPhi[j]) / vol;    

}

//
void assembleGlobalStiffMatrix ( const TriangMesh& triangMesh, SparseMatrix& StiffMat ) {
  if( triangMesh.getNumberOfNodes() != StiffMat.getRows() )
    throw "assembleGlobalStiffMatrix::wrong number of rows!";
  if( triangMesh.getNumberOfNodes() != StiffMat.getCols() )
    throw "assembleGlobalStiffMatrix::wrong number of rows!";
  
  // set zero
  StiffMat.setZero();
  
  // run over all faces
  for( int i = 0; i < triangMesh.getNumberOfFaces(); i++ ){
    
    // compute local matrix
    double LocalMatrix[3][3];
    assembleLocalStiffMatrix ( triangMesh, i, LocalMatrix );
    
    // insert in global matrix
    for( int j = 0; j < 3; j++ )
      for( int k = 0; k < 3; k++ )
	StiffMat( triangMesh.getGlobalIdx(i,j), triangMesh.getGlobalIdx(i,k) ) += LocalMatrix[j][k];    
  }
}


// computes (\int f \phi^h_i dx)_i, where \phi^h_i is the linear FEM basis function located at node i
void integrateOnePointQuadratureRule( const TriangMesh& mesh, const Vector& f, Vector& integral ) {
  if( f.size() != mesh.getNumberOfNodes() )
    throw "integrateOnePointQuadratureRule: input vector has wrong size!";
  if( f.size() != integral.size() )
    throw "integrateOnePointQuadratureRule: output vector has wrong size!";
  
  integral.setZero();
  // phi^h_i(c) = 1/3 if c is the center point of an arbitrary triangle
  for( int i = 0; i < mesh.getNumberOfFaces(); i++ ){
    double vol = mesh.getArea ( i );
    for( int j = 0; j < 3; j++ )
      integral[ mesh.getGlobalIdx(i,j) ] += vol * f[ mesh.getGlobalIdx(i,j) ] / 3.;    
  }
}

//
void setZeroBoundaryConditions( const TriangMesh& mesh, SparseMatrix& SystemMat, Vector& rhs ) {
  for( int i = 0; i < mesh.getNumberOfNodes(); i++ ){
    if( mesh.isBoundaryNode(i) ){
      rhs[i] = 0.;
      SystemMat.setRowZero( i );
      for( int j = 0; j < mesh.getNumberOfNodes(); j++ )
        SystemMat.set( j, i, 0. );
      SystemMat.set( i, i, 1.0 );
    }
  }
}
  
//=========================================================
//=========================================================

// Hauptprogramm
int main () {
    
  try{      
  
    clock_t begin, end;
    
    // load triangular mesh
    TriangMesh mesh;
    mesh.loadFromOBJ ( "square.obj" );
    cerr << mesh.getNumberOfNodes() << " " << mesh.getNumberOfFaces() << endl;
    
    // assemble stiffness matrix
    SparseMatrix StiffMat( mesh.getNumberOfNodes(), mesh.getNumberOfNodes() );
    assembleGlobalStiffMatrix ( mesh, StiffMat );
    //StiffMat.print();
    
    // define vectors
    Vector sol( mesh.getNumberOfNodes() ), rhs( mesh.getNumberOfNodes());
    sol.setAll( 1. );
    
    // set up rhs for f = 1 and plot
    integrateOnePointQuadratureRule( mesh, sol, rhs );
    plotPNG( mesh, rhs, "rhs.png" );
    
    // set zero boundary conditions
    setZeroBoundaryConditions( mesh, StiffMat, rhs );
    
    // set up solver
    CGSolver cgSolver( 1e-8, 10000 );

    // solve 
    begin = clock();     
    cgSolver.solve( StiffMat, rhs, sol );    
    end = clock();
  
    cerr << "Time for CG = " << double(end - begin) / CLOCKS_PER_SEC << endl;
    
    // plot solution
    plotPNG( mesh, sol, "sol.png" );
  
  }
  catch ( const char* exc ){
    cerr << "Exception caught: " << exc << endl;
  }

  return 0;
}